'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global log, getUnicodeDomain, DomainSet, setDefault, parseUri,
loggingEnable, blockCounts, createMalwareNotification, MaxCacheEntries */

// Filter objects representing the given filter text.
function FilterSet() {
  // Map from domain (e.g. 'mail.google.com', 'google.com', or special-case
  // 'global') to list of filters that specify inclusion on that domain.
  // E.g. /f/$domain=sub.foo.com,bar.com will appear in items['sub.foo.com']
  // and items['bar.com'].
  this.items = { global: [] };

  // Map from domain to set of filter ids that specify exclusion on that domain.
  // Each filter will also appear in this.items at least once.
  // Examples:
  //   /f/$domain=~foo.com,~bar.com would appear in
  //     items['global'], exclude['foo.com'], exclude['bar.com']
  //   /f/$domain=foo.com,~sub.foo.com would appear in
  //     items['foo.com'], exclude['sub.foo.com']
  this.exclude = {};
}

// Construct a FilterSet from the Filters that are the values in the |data|
// object.  All filters should be the same type (whitelisting PatternFilters,
// blocking PatternFilters, or SelectorFilters.)
FilterSet.fromFilters = function filterSetFromFilters(data) {
  const result = new FilterSet();

  for (const aKey in data) {
    const filter = data[aKey];

    for (const d in filter.domains.has) {
      if (filter.domains.has[d]) {
        const key = (d === DomainSet.ALL ? 'global' : d);
        setDefault(result.items, key, []).push(filter);
      } else if (d !== DomainSet.ALL) {
        setDefault(result.exclude, d, {})[filter.id] = true;
      }
    }
  }

  return result;
};

FilterSet.prototype = {

  // Return a new FilterSet containing the subset of this FilterSet's entries
  // which relate to the given domain or any of its superdomains.  E.g.
  // sub.foo.com will get items['global', 'foo.com', 'sub.foo.com'] and
  // exclude['foo.com', 'sub.foo.com'].
  viewFor(domain, matchGeneric) {
    const result = new FilterSet();
    if (!matchGeneric) {
      result.items.global = this.items.global;
    }

    for (const nextDomain in DomainSet.domainAndParents(domain)) {
      if (this.items[nextDomain]) {
        result.items[nextDomain] = this.items[nextDomain];
      }
      if (this.exclude[nextDomain]) {
        result.exclude[nextDomain] = this.exclude[nextDomain];
      }
    }

    return result;
  },

  // Get a list of all Filter objects that should be tested on the given
  // domain, and return it with the given map function applied. This function
  // is for hiding rules only
  filtersFor(domain, matchGeneric) {
    const unicodeDomain = getUnicodeDomain(domain);
    const limited = this.viewFor(unicodeDomain, matchGeneric);
    const data = {};

    // data = set(limited.items)
    for (const subdomain in limited.items) {
      const entry = limited.items[subdomain];
      for (let i = 0; i < entry.length; i += 1) {
        const filter = entry[i];
        data[filter.id] = filter;
      }
    }

    // data -= limited.exclude
    for (const subdomain in limited.exclude) {
      for (const filterId in limited.exclude[subdomain]) {
        delete data[filterId];
      }
    }

    const result = [];
    for (const k in data) {
      result.push(data[k].selector);
    }
    return result;
  },

  // Get a list of all Filter objects that should be tested on the given
  // domain, and return it with the given map function applied. This function
  // is for advanced hiding rules only
  advanceFiltersFor(domain, matchGeneric) {
    const unicodeDomain = getUnicodeDomain(domain);
    const limited = this.viewFor(unicodeDomain, matchGeneric);
    const data = {};

    // data = set(limited.items)
    for (const subdomain in limited.items) {
      const entry = limited.items[subdomain];
      for (let i = 0; i < entry.length; i += 1) {
        const filter = entry[i];
        data[filter.id] = filter;
      }
    }

    // data -= limited.exclude
    for (const subdomain in limited.exclude) {
      for (const filterId in limited.exclude[subdomain]) {
        delete data[filterId];
      }
    }

    const result = [];
    for (const k in data) {
      result.push(data[k]);
    }
    return result;
  },

  // Return the filter that matches this url+elementType on this frameDomain:
  // the filter in a relevant entry in this.items who is not also in a
  // relevant entry in this.exclude.
  // isThirdParty: true if url and frameDomain have different origins.
  matches(url, elementType, frameDomain, isThirdParty, matchGeneric) {
    const limited = this.viewFor(frameDomain, matchGeneric);
    for (const k in limited.items) {
      const entry = limited.items[k];
      for (let i = 0; i < entry.length; i += 1) {
        const filter = entry[i];
        if (filter.matches(url, elementType, isThirdParty)) {
          // Maybe filter shouldn't match because it is excluded on our domain?
          let excluded = false;
          for (const k2 in limited.exclude) {
            if (limited.exclude[k2][filter.id]) {
              excluded = true;
              break;
            }
          }

          if (!excluded) {
            return filter;
          }
        }
      }
    }

    return null;
  },
};

const BlockingFilterSet = function (patternFilterSet, whitelistFilterSet) {
  this.pattern = patternFilterSet;
  this.whitelist = whitelistFilterSet;
  // Caches results for this.matches()
  this.matchCache = {};
  this.numCacheEntries = 0;
};

// Checks if the two domains have the same origin
// Inputs: the two domains
// Returns: true if third-party, false otherwise
BlockingFilterSet.checkThirdParty = function checkIfThirdParty(domain1, domain2) {
  const match1 = parseUri.secondLevelDomainOnly(domain1, false);
  const match2 = parseUri.secondLevelDomainOnly(domain2, false);
  return (match1 !== match2);
};

BlockingFilterSet.prototype = {

  // True if the url is blocked by this filterset.
  // Inputs:
  //   url:string - The URL of the resource to possibly block
  //   elementType:ElementType - the type of element that is requesting the
  //                             resource
  //   frameDomain:string - domain of the frame on which the element resides
  //   returnFilter?:bool - see Returns
  //   returnTuple?:bool - see Returns
  // Returns:
  //   if returnFilter is true:
  //       text of matching pattern/whitelist filter, null if no match
  //   if returnFilter is false:
  //       true if the resource should be blocked, false otherwise
  //   if returnTuple is true and returnFilter is true:
  //       returns an object containing two properties:
  //          'blocked' - true or false
  //          'text' - text of matching pattern/whitelist filter, null if no match
  matches(url, elementType, frameDomain, returnFilter, returnTuple) {
    const urlDomain = getUnicodeDomain(parseUri(url).hostname);
    const isThirdParty = BlockingFilterSet.checkThirdParty(urlDomain, frameDomain);

    // matchCache approach taken from ABP
    const key = `${url} ${elementType} ${isThirdParty}`;
    if (key in this.matchCache) {
      return this.matchCache[key];
    }

    let match = this.whitelist.matches(url, elementType, frameDomain, isThirdParty);
    if (match) {
      if (loggingEnable) {
        log(frameDomain, ': whitelist rule', match.rule, 'exempts url', url);
      }

      this.checkCacheSize();
      if (returnTuple && returnFilter) {
        this.matchCache[key] = { blocked: false, text: match.text };
      } else {
        this.matchCache[key] = (returnFilter ? match.text : false);
      }

      this.numCacheEntries += 1;
      return this.matchCache[key];
    }

    match = this.pattern.matches(url, elementType, frameDomain, isThirdParty);
    if (match) {
      if (loggingEnable) {
        log(frameDomain, ': matched', match.rule, 'to url', url);
      }

      this.checkCacheSize();
      if (returnTuple && returnFilter) {
        this.matchCache[key] = { blocked: true, text: match.text };
      } else {
        this.matchCache[key] = (returnFilter ? match.text : true);
      }

      this.numCacheEntries += 1;
      return this.matchCache[key];
    }

    if (this.malwareDomains &&
        urlDomain &&
        this.malwareDomains[urlDomain.charAt(0)] &&
        this.malwareDomains[urlDomain.charAt(0)].indexOf(urlDomain) > -1) {
      if (loggingEnable) {
        log('matched domain', urlDomain);
      }

      if (blockCounts) {
        blockCounts.recordOneMalwareBlocked();
      }

      this.checkCacheSize();
      this.matchCache[key] = (returnFilter ? urlDomain : true);

      // createMalwareNotification is not defined outside of BG page
      if (typeof createMalwareNotification === 'function') {
        createMalwareNotification(frameDomain);
      }

      this.numCacheEntries += 1;
      return this.matchCache[key];
    }

    this.checkCacheSize();
    this.matchCache[key] = false;
    this.numCacheEntries += 1;
    return this.matchCache[key];
  },

  setMalwareDomains(malwareDoms) {
    if (malwareDoms === null) {
      this.malwareDomains = null;
      return;
    }

    const domains = malwareDoms.adware;
    const result = {};
    for (let i = 0; i < domains.length; i += 1) {
      const domain = domains[i];
      const char = domain.charAt(0);
      if (!result[char]) {
        result[char] = [];
      }

      result[char].push(domain);
    }

    this.malwareDomains = result;
  },

  getMalwareDomains() {
    return this.malwareDomains;
  },

  checkCacheSize() {
    if (this.numCacheEntries >= MaxCacheEntries) {
      this.resetMatchCache();
    }
  },

  resetMatchCache() {
    this.matchCache = {};
    this.numCacheEntries = 0;
  },
};

// Attach object to window
Object.assign(window, {
  BlockingFilterSet,
});
